/*
** commands.c - handles user commands
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

/* includes */
#include "ipxfury.h"
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


/* the commands we know */

#define TCMD_QUIT		1
#define TCMD_FURY		2
#define TCMD_TERM		3
#define TCMD_INTERFACES		4
#define TCMD_EVENTS		5
#define TCMD_HWADDR		6
#define TCMD_HDRCACHE		7
#define TCMD_MODEM		8
#define TCMD_PKTCACHE		9
#define TCMD_HELP		10
#define TCMD_PORTS		11
#define TCMD_WATCHPORT		12
#define TCMD_IGNOREPORT		13
#define TCMD_RESET		14
#define TCMD_STATS		15
#define TCMD_UDP		16
#define TCMD_VERBOSE		17
#define TCMD_SET		18
#define TCMD_OPTIONS		19
#define TCMD_VERSION		20

COMMAND_ENTRY	cmd_entries[] = {
	{ "quit",		TCMD_QUIT },
	{ "fury",		TCMD_FURY },
	{ "term",		TCMD_TERM },
	{ "interfaces",		TCMD_INTERFACES },
	{ "events",		TCMD_EVENTS },
	{ "hwaddr",		TCMD_HWADDR },
	{ "hdrcache",		TCMD_HDRCACHE },
	{ "modem",		TCMD_MODEM },
	{ "pktcache",		TCMD_PKTCACHE },
	{ "help",		TCMD_HELP },
	{ "?",			TCMD_HELP },
	{ "ports",		TCMD_PORTS },
	{ "watchport",		TCMD_WATCHPORT },
	{ "ignoreport",		TCMD_IGNOREPORT },
	{ "reset",		TCMD_RESET },
	{ "stats",		TCMD_STATS },
	{ "udp",		TCMD_UDP },
	{ "verbose",		TCMD_VERBOSE },
	{ "set",		TCMD_SET },
	{ "options",		TCMD_OPTIONS },
	{ "version",		TCMD_VERSION },
	{ "",			-1 },		/* tag ending */
};

#define UCMD_CONNECT		1
#define UCMD_THRESHOLD		2
#define UCMD_MAP		3
#define UCMD_IPS		4
#define UCMD_PKT		5
#define UCMD_DATA		6

COMMAND_ENTRY udp_cmds[] = {
	{ "connect",		UCMD_CONNECT },
	{ "threshold",		UCMD_THRESHOLD },
	{ "map",		UCMD_MAP },
	{ "ips",		UCMD_IPS },
	{ "pkt",		UCMD_PKT },
	{ "data",		UCMD_DATA },
	{ "",			-1 },		/* end marker */
};


/* globals */
char *help_str =	"fury                           - starts the fury\n"
			"term [interface]               - talk to the modem\n"
			"interfaces                     - shows the interfaces\n"
			"events                         - shows schedules events\n"
			"hwaddr                         - lists the known hardware addresses\n"
			"hdrcache interface [s/r]       - shows the send or receive header cache\n"
			"modem [interface]              - shows the modem controls of a serial device\n"
			"pktcache interface [s/r]       - shows the send or receive data cache\n"
			"stats [time]                   - shows device stats and changes \n"
			"ports                          - shows the ports (IPX) we are watching\n"
			"ignoreport [all/num] [num]...  - ignores that port (IPX)\n"
			"watchport [all/num] [num]...   - watches that port (IPX)\n"
			"udp                            - controls the UDP interface (try: udp help)\n"
			"verbose [level]                - sets the verbose level\n"
			"set [interface] [opt] [on/off] - turns the option on/off for that interface\n"
			"options [interface]            - lists the options for that interface\n"
			"quit                           - exit the FURY!";
			

char *udp_help_str =	"-= UDP COMMANDS =-\n"
			"udp connect [hostname] [port]      - spreads the FURY!\n"
			"udp map                            - interesting to the insane...\n"
			"udp ips                            - list the furious crowd!\n" 
			"udp pkt                            - show packet caches\n"
			"udp data                           - show data caches\n"
			"udp threshold [bytes] [q] [genmax] - high tech!";


char *opts_str =	"-= INTERFACE OPTIONS =-\n"
			"fullframe                       - sends a full fury frame first\n"
			"forward                         - forwards broadcast packets\n"
			"sc                              - handle starcraft data better\n"
			"frametype                       - changes the ethernet frame type";

/* externs */
extern GLOBALS	ifg;

/* protos */

/*
**	PROCESS_TERMINAL_COMMAND - handles user command issues
*/
void process_terminal_command(char *cmd_string)
{
	char	*token[2];
	
	
	/* clean it */
	cleanline(cmd_string);
	
	/* is this a good string */
	if( strlen(cmd_string) == 0 )
		return;
		
	/* get cmd and args */
	tokenize(cmd_string, 2, ' ', token);
	
	/* now do the command */
	switch( which_command(cmd_entries, token[0]) ) {
		case CMD_UNKNOWN:
			notice("Unknown command.\n");
			break;
		case CMD_AMBIGIOUS:
			notice("Ambigious command.\n");
			break;
			
		case TCMD_QUIT:			/* -- QUIT -- */
			if( ifg.status != FSTATUS_PROMPT ) {
				notice("Use ^D or ^C to quit.\n");
			} else {
				ifg.done = 1;
			}
			break;
			
		case TCMD_FURY:			/* -- FURY -- */
			if( ifg.status != FSTATUS_FURIOUS ) {
				/* little notice */
				notice("((( Fury Unleashed! [^D to quit] )))\n");
				ipx_fury();
			} else {
				notice("What is this?! You want to be even MORE FURIOUS?!?\n");
			}
			break;
		
		case TCMD_TERM:			/* -- TERM -- */
			if( ifg.status != FSTATUS_FURIOUS ) {
				fury_term(token[1]);
			}
			break;
			
		case TCMD_INTERFACES:		/* -- INTERFACES -- */
			interfaces_info();
			break;
			
		case TCMD_EVENTS:		/* -- EVENTS -- */
			notice("Stuff to do:\n");
			event_dump(ifg.events);
			break;
			
		case TCMD_HWADDR:		/* -- HWADDR -- */
			notice("Hardware Address Hash Table (%d entries)\n", ifg.hwaddr_table.entries);
			hash_enum_entries(&ifg.hwaddr_table, (void (*)(int, void*))ether_hash_enum);	
			break;
			
		case TCMD_HDRCACHE:		/* -- HDRCACHE -- */
			show_hdrcache(token[1]);
			break;
			
		case TCMD_MODEM:		/* -- MODEM -- */
			show_modem_info(token[1]);
			break;
			
		case TCMD_PKTCACHE:		/* -- PKTCACHE -- */
			show_pktcache(token[1]);
			break;
			
		case TCMD_HELP:			/* -- HELP -- */
			show_help();
			break;
			
		case TCMD_PORTS:		/* -- PORTS -- */
			notice("Ports we like:\n");
			show_set_ports(ifg.watched_ports);
			break;
			
		case TCMD_WATCHPORT:		/* -- WATCHPORT -- */
			command_changeports(1, token[1]);
			break;
			
		case TCMD_IGNOREPORT:		/* -- IGNOREPORT -- */
			command_changeports(0, token[1]);
			break;

		case TCMD_RESET:		/* -- RESET -- */
			command_reset(1);
			break;			
			
		case TCMD_STATS:		/* -- STATS -- */
			show_stats(token[1]);
			break;		
			
		case TCMD_UDP:			/* -- UDP -- */
			udp_command(token[1]);
			break;
			
		case TCMD_VERBOSE:		/* -- VERBOSE -- */
			command_verbose(token[1]);
			break;
		
		case TCMD_SET:			/* -- SET -- */
			command_set(token[1]);
			break;	
		
		case TCMD_OPTIONS:
			command_options(token[1]);
			break;
			
		case TCMD_VERSION:		/* -- VERSION -- */
			notice("IPXFury version: %d.%d.%d\n", ifg.ver_major, ifg.ver_minor, ifg.ver_build);
			break;
			
	}			

	/* done */
}

char *on_off(int val)
{
	if( val )
		return("on");
	else
		return("off");
}

/*
**	OPTIONS - shows 'em
*/
void command_options(char *args)
{
	INTERFACE	*iface, *wanted;
	int		i;
	
	/* do they have args? */
	if( args != NULL ) {
		/* get the interface */
		wanted = interface_find_name(args);
		if( NULL == wanted ) {
			/* error */
			notice("Could not find interface '%s'\n", args);
			return;
		}
	} else {
		/* no args.. wanted is NULL */
		wanted = NULL;
	}
	
	/* for each interface */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		/* set ptr */
		iface = &ifg.interfaces[i];
		
		/* is this our wanted */
		if( (iface == wanted) || (wanted == NULL) ) {
			/* show it */
			notice("Interface [%-8s]:\n", iface->name);
			
			/* some stuff */
			if( iface->type == FURY_IF_TYPE_ETHER ) {
				notice("\tStarcraft Optimizations .......... %s\n", on_off(iface->options & FOPT_SC_RESEND));
				notice("\tFrame Type ....................... ");
				switch( iface->data.e.e_frametype ) {
					case FURY_FRAMETYPE_II:
						notice("Ethernet II\n");	break;
					case FURY_FRAMETYPE_802_3:
						notice("802.3\n");	break;
					case FURY_FRAMETYPE_802_2:
						notice("802.2\n");	break;
					default:
						notice("UNKNOWN!\n");	break;
				}
			} else if( iface->type == FURY_IF_TYPE_SERIAL ) {
				notice("\tNo Available Options.\n");
			} else if( iface->type == FURY_IF_TYPE_UDP ) {
				notice("\tFull Frame ....................... %s\n", on_off(iface->options & FOPT_FULLFRAME));
				notice("\tBroadcast Forwarding ............. %s\n", on_off(iface->options & FOPT_FORWARD));
				notice("\tMinimum Packet Size .............. %d\n", iface->data.u.u_buffer_threshold);
				notice("\tMaximum Wait ..................... %d (ms)\n", ((1000 * iface->data.u.u_buffer_wait_sec) + (iface->data.u.u_buffer_wait_usec / 1000)));
				notice("\tDiff Generation Maximum .......... %d\n", iface->generation_max);
			}			
		}
	}
	
	/* done */
	return;
		
		
}

/*
**	SET_OPTIONS - this actually does the setting
*/
int set_option(INTERFACE *iface, int option, int type, char *value, char *opt_str)
{
	int		set, frametype;
	int		fopt;
	
	/* is this a boolean type? */
	if( type == OPT_TYPE_BOOL ) {
		/* do they want to set it or not? */
		if( strcasecmp(value, "on") == 0 ) {
			set = 1;
		} else if( strcasecmp(value, "off") == 0 ) {
			set = 0;
		} else {
			/* error */
			notice("Use 'on' or 'off' for option '%s'\n", opt_str);
			return(0);
		}
		
		/* now set the fury option */
		switch( option ) {
			case OPT_FULLFRAME:
				fopt = FOPT_FULLFRAME;
				break;
			case OPT_FORWARD:
				fopt = FOPT_FORWARD;
				break;
			case OPT_SC_OPTS:
				fopt = FOPT_SC_RESEND;
				break;
			default:
				errormsg(0, "set_option", "bad bool option!");
				return(0);
		}
		
		/* if it on or off */
		if( set ) {
			iface->options |= fopt;
		} else {
			iface->options &= ~fopt;
		}
		
		/* message */
		notice("Interface %s: Option: %s (%s)\n", iface->name, opt_str, value);
			
	} else if( type == OPT_TYPE_VALUE ) {
		/* do something based on the option */
		switch( option ) {
			case OPT_ETH_FRAMETYPE:
				/* make sure this is an ethernet interface */
				if( iface->type == FURY_IF_TYPE_ETHER ) {
					/* check args */
					if( strcasecmp(value, "ii") == 0 ) {
						frametype = FURY_FRAMETYPE_II;
					} else if( strcasecmp(value, "802.3") == 0 ) {
						frametype = FURY_FRAMETYPE_802_3;
					} else if( strcasecmp(value, "802.2") == 0 ) {
						frametype = FURY_FRAMETYPE_802_2;
					} else {
						/* punk! */
						notice("Unknown frametype '%s'!\n", value);
						return(0);
					}
					/* change it */
					if( ether_changeframe(iface, frametype) ) {
						/* msg */
						notice("Changed frametype to '%s' on '%s'\n", value, iface->name);
					}
				} else {
					/* stoopid */
					notice("You can only select frametypes on ethernet interfaces!\n");
					return(0);
				}
				break;
			
			default:
				errormsg(0, "set_option", "bad value option!");
				return(0);
		}
	
	}

	/* ok */
	return(1);
}


/*
**	SET -
**
**	set [interface] [option] [on/off]
*/
void command_set(char *args) 
{
	char		*token[4];
	INTERFACE	*iface;
	int		option, type;
	
	/* token it up */
	if( 3 != tokenize(args, 4, ' ', token) ) {
		/* usage! */
		notice("Usage: set [interface] [option] [on/off]\n");
		notice("%s\n", opts_str);
		return;
	}
	
	/* now the options */
	if( strcasecmp(token[1], "fullframe") == 0 ) {
		option = OPT_FULLFRAME;
		type = OPT_TYPE_BOOL;
	} else if( strcasecmp(token[1], "forward") == 0 ) {
		option = OPT_FORWARD;
		type = OPT_TYPE_BOOL;
	} else if( strcasecmp(token[1], "sc") == 0 ) {
		option = OPT_SC_OPTS;
		type = OPT_TYPE_BOOL;
	} else if( strcasecmp(token[1], "frametype") == 0 ) {
		option = OPT_ETH_FRAMETYPE;
		type = OPT_TYPE_VALUE;
	} else {
		/* what? */
		notice("Unknown option '%s'\n", token[1]);
		notice("%s\n", opts_str);
		return;
	}
	
	/* get the interface */
	iface = interface_find_name(token[0]);
	if( NULL == iface ) {
		/* problem */
		notice("Could not find interface '%s'\n", token[0]);
		return;
	}
	
	/* make the call */
	set_option(iface, option, type, token[2], token[1]);
	
	/* done */
	return;	

}


/*
**	VERBOSE
*/
void command_verbose(char *arg)
{
	/* if args.. set the new one */
	if( arg != NULL ) {
		ifg.verbose_level = strtol(arg, NULL, 0);
		
		/* limit check it */
		if( ifg.verbose_level < 0 ) {	
			ifg.verbose_level = 0;
		} else if( ifg.verbose_level > 10 ) {
			ifg.verbose_level = 10;
		}
	}
	
	/* show it */
	notice("Current verbosity level: %2d (0..10)\n", ifg.verbose_level);

	/* done */

}

/*
**	UDP_COMMAND -
*/ 
void udp_command(char *in_args)
{
	struct sockaddr_in	sin;
	int			portnum;
	INTERFACE		*iface;
	char			*token[5];
	int			ms;
			
	/* do they have a UDP interface? */
	iface = interface_find_name("udp");
	if( NULL == iface ) {
		notice("There is no UDP interface!!\n");
		return;
	}
	
	/* do we have args? */
	if( in_args == NULL ) {
		notice("%s\n", udp_help_str);
		return;	
	}
	
	/* get args */
	tokenize(in_args, 5, ' ', token);

	/* what command */	
	switch( which_command(udp_cmds, token[0]) ) {
		case CMD_UNKNOWN:
			notice("Unknown UDP command '%s'.\n", token[0]);
			notice("%s\n", udp_help_str);
			break;
		case CMD_AMBIGIOUS:
			notice("Ambigious UDP command.\n");
			break;
		case UCMD_CONNECT:
			/* do we have args? */
			if( (token[1] == NULL) || (token[2] == NULL) ) {
				notice("Usage: connect [hostname] [port]\n");
			} else {
				/* only when furious */
				if( ifg.status != FSTATUS_FURIOUS ) {
					notice("[udp] You can only connect while furious\n");
					return;
				}
				
				/* get port number */
				portnum = strtol(token[2], NULL, 0);
				
				/* clean */
				memset(&sin, 0, sizeof(sin));
				sin.sin_family 	= AF_INET;
				sin.sin_port	= htons(portnum);
				
				/* we are going to try and connect to this guy! */
				if( get_ipaddr(token[1], &sin.sin_addr) ) {
					/* ok, we have the address .. lets send a hello */
					udp_connect_req(iface, &sin);
					notice("[udp] Sent connect request to %s:%d...\n", token[1], portnum);
				} else {
					notice("[udp] Unknown hostname %s\n", token[1]);
				}
			}
			break;
			
		case UCMD_THRESHOLD:
			/* do they want to set or show? */
			if( (token[1] != NULL) && (token[2] != NULL) && (token[3] != NULL) ) {
				/* set it */
				iface->data.u.u_buffer_threshold = strtol(token[1], NULL, 0);
				ms = strtol(token[2], NULL, 0);
				iface->data.u.u_buffer_wait_sec	 = ms / 1000;
				iface->data.u.u_buffer_wait_usec = (ms % 1000) * 1000;
				iface->generation_max		 = strtol(token[3], NULL, 0);
			}
			/* show it */
			notice("[udp] Min packet size (%5d bytes)  Max wait (%5dms)  Gen Max (%5d)\n", iface->data.u.u_buffer_threshold, ((1000 * iface->data.u.u_buffer_wait_sec) + (iface->data.u.u_buffer_wait_usec / 1000)), iface->generation_max);
			break;
		
		case UCMD_MAP:
			/* do it */
			notice("-- Hardware Address to IP Address Map (%d entries) --\n", iface->data.u.u_hw_to_ip.entries);
			notice("Num  Hardware Addr        IP Addr\n");
			hash_enum_entries(&iface->data.u.u_hw_to_ip, (void (*)(int, void *))udp_dump_hwtable);
			break;
		
		case UCMD_IPS:
			/* do it some more */
			notice("-- List of IP's We Are Connected To (%d entries) --\n", iface->data.u.u_ipaddrs.entries);
			notice("Num  IP Address      Port  Hostname                  Ver       Sent    Recv\n");           
			    /*  [..] 111.222.333.444 12345 1234567890123456789012345 00.00.000 999.98K 999.98K*/
			hash_enum_entries(&iface->data.u.u_ipaddrs, (void (*)(int, void *))udp_dump_iptable);				
			break;
		
	}			
	
	/* done */
	return; 
}



/*
**	SHOW_STATS
*/
void show_stats(char *args)
{
	int		i;
	int		hit_percent;
	int		miss_percent;
	int		total;

	/* if they gave an argument, use it as the recent stats time */
	if( args != NULL ) {
		/* set the time */
		ifg.stats.recent_stats_time = strtol(args, NULL, 0);
		
		/* show them */
		notice("Recent Statistics Interval (%d:%.2d)\n", (ifg.stats.recent_stats_time / 60), (ifg.stats.recent_stats_time % 60));
		
		/* remove the event */
		remove_events(ifg.events, ETYPE_PERIODIC_STATS, NULL);
		
		/* add it in */
		add_event(ifg.events,
			  ETYPE_PERIODIC_STATS,
			  ifg.stats.recent_stats_time,
			  ifg.stats.recent_stats_time,
			  NULL, NULL);		
		
		/* done */
		return;
	}

	/* for each interfaces, show the info */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		notice("%s Stats:\n", ifg.interfaces[i].name);
		total = ifg.interfaces[i].stats.hdrcache_hits + ifg.interfaces[i].stats.hdrcache_misses;
		if( total > 0 ) {
			hit_percent 	= (100 * ifg.interfaces[i].stats.hdrcache_hits) / total;
			miss_percent	= (100 * ifg.interfaces[i].stats.hdrcache_misses) / total;
			notice("\tHeader Cache  Hits: %8d (%3d%%)  Misses: %8d (%3d%%)\n", ifg.interfaces[i].stats.hdrcache_hits, hit_percent, ifg.interfaces[i].stats.hdrcache_misses, miss_percent);
		}
		
		total = ifg.interfaces[i].stats.pktcache_hits + ifg.interfaces[i].stats.pktcache_misses;
		if( total > 0 ) {
			hit_percent 	= (100 * ifg.interfaces[i].stats.pktcache_hits) / total;
			miss_percent	= (100 * ifg.interfaces[i].stats.pktcache_misses) / total;
			notice("\tData Cache    Hits: %8d (%3d%%)  Misses: %8d (%3d%%)\n", ifg.interfaces[i].stats.pktcache_hits, hit_percent, ifg.interfaces[i].stats.pktcache_misses, miss_percent);
		}
		
		/* show bytes sent and received */
		notice("\tTotal Bytes   Sent: %8s           ", print_datatrans(ifg.interfaces[i].stats.bytes_sent));
		notice("Rcvd: %8s\n", print_datatrans(ifg.interfaces[i].stats.bytes_rcvd));
		

	}

	/* general stats */
	notice("General Stats:\n");
	/* show some info */
	notice("\tOne Byte Packets Sent:  %d\n", ifg.stats.one_byte_packets);
	notice("\tRecent Stats Interval:  %d:%.2d \n", (ifg.stats.recent_stats_time / 60), (ifg.stats.recent_stats_time % 60));

		
	/* done */

}



/* 
**	COMMAND_RESET
*/
void command_reset(int send_reset)
{

	/* this is broken */
	notice("THIS IS BROKEN!!!\n");
	return;
}



/*
**	COMMAND_CHANGEPORTS
*/
void command_changeports(int set, char *args)
{
	int		portnum;
	char		*token[2];

	/* give them usage */
	if( args == NULL ) {
		if( set )
			notice("Usage: watchport port port port...\n");
		else
			notice("Usage: ignoreport port port port..\n");
		return;
	}
	
	/* for each piece do it up */
	do {
		/* get the 1st one */
		tokenize(args, 2, ' ', token);
		args = token[1];
		if( token[0] != NULL ) {
			if( strcasecmp(token[0], "all") == 0 ) {
				/* change them all */
				if( set ) 
					set_all_ports(ifg.watched_ports);
				else
					unset_all_ports(ifg.watched_ports);
			} else {
				/* change only 1 */
				portnum = strtol(token[0], NULL, 0);
				if( (portnum < 0) || (portnum > 0xFFFF) ) {
					notice("Invalid port %d (%.4x)\n", portnum, portnum);
				} else {
					if( set )
						set_port(ifg.watched_ports, portnum);
					else
						unset_port(ifg.watched_ports, portnum);
				}
			}
		}
	} while( token[0] != NULL );
	
	/* done */
	notice("Changed port table.\n");
}


/*
**	SHOW_HELP
*/ 
void show_help()
{

	notice("-= IPX FURY COMMANDS =-\n");
	notice("%s\n", help_str);

}



/*
**	SHOW_PKTCACHE
*/
void show_pktcache(char *args)
{
	INTERFACE	*iface;
	char		*token[2];

	/* need args */
	if( args == NULL ) {
		notice("Usage: pktcache [interface name] [s/r]\n");
		return;
	}

	/* get interface */
	tokenize(args, 2, ' ', token);
	
	iface = interface_find_name(token[0]);
	if( iface == NULL ) {
		errormsg(0, "show_pktcache", "could not find interface '%s'", token[0]);
		return;
	}

	/* so show or recv? */
	if( (token[1] == NULL) || (*token[1] == 's') )	{
		/* got one, show it's info */
		notice("-- Send Packet Cache for Interface [%s] --\n", iface->name);
		if( iface->cache_ptrs.send_pktcache != NULL ) 
			hash_enum_entries(iface->cache_ptrs.send_pktcache, (void (*)(int, void *))pktcache_hash_enum);
		else
			notice("No Cache!\n");
	} else if( *token[1] == 'r' ) {
		notice("-- Receive Packet Cache for Interface [%s] --\n", iface->name);
		if( iface->cache_ptrs.recv_pktcache != NULL ) 
			hash_enum_entries(iface->cache_ptrs.recv_pktcache, (void (*)(int, void *))pktcache_hash_enum);
		else
			notice("No Cache!\n");
	} else {
		notice("Unknown mode '%s'\n", token[1]);
	}


}




/*
**	SHOW_MODEM_INFO - shows info about the modem
*/
void show_modem_info(char *args)
{
	INTERFACE	*iface;
	int		modem_lines, i;
	
	if( args == NULL ) {
		/* show every serial device */
		i = 1;
		do {
			iface = interface_find(i, FURY_IF_TYPE_SERIAL);
			if( iface != NULL ) {
				/* now get the stats */
				if( serial_status(iface, &modem_lines) ) {
					notice("%s Status:", iface->name);
					if( modem_lines & TIOCM_DTR )
						notice(" DTR");
					if( modem_lines & TIOCM_RTS )
						notice(" RTS");
					if( modem_lines & TIOCM_DSR )
						notice(" DSR");
					if( modem_lines & TIOCM_CTS )
						notice(" CTS");
					if( modem_lines & TIOCM_CAR )
						notice(" CAR");
					if( modem_lines & TIOCM_RNG )
						notice(" RNG");
					if( modem_lines & TIOCM_ST )
						notice(" ST");
					if( modem_lines & TIOCM_SR )
						notice(" SR");
					notice("\n");
				} else {
					errormsg(0, "show_modem_inf", "could not get status of '%s'", iface->name);	
				}
				
				/* next one */
				i++;
			}
		} while( iface != NULL );
		
		/* did they even show one? */
		if( i == 1 ) {
			errormsg(0, "show_modem_info", "could not find a serial device");
			return;
		}
		
	} else {
		/* show the one they want */
		iface = interface_find_name(args);
		if( iface == NULL ) {
			errormsg(0, "show_modem_info", "no such interface '%s'", args);
			return;
		}
		
		/* show the info */
		/* now get the stats */
		if( serial_status(iface, &modem_lines) ) {
			notice("%s Status:", iface->name);
			if( modem_lines & TIOCM_DTR )
				notice(" DTR");
			if( modem_lines & TIOCM_RTS )
				notice(" RTS");
			if( modem_lines & TIOCM_DSR )
				notice(" DSR");
			if( modem_lines & TIOCM_CTS )
				notice(" CTS");
			if( modem_lines & TIOCM_CAR )
				notice(" CAR");
			if( modem_lines & TIOCM_RNG )
				notice(" RNG");
			if( modem_lines & TIOCM_ST )
				notice(" ST");
			if( modem_lines & TIOCM_SR )
				notice(" SR");
			notice("\n");
		} else {
			errormsg(0, "show_modem_inf", "could not get status of '%s'", iface->name);
		}
		
	}



}



/*
**	SHOW_HDRCACHE - shows the cache for a device
*/
void show_hdrcache(char *args)
{
	INTERFACE		*iface;
	char			*token[2];

	/* any args */
	if( args == NULL ) {
		notice("Usage: hdrcache [interface name] [s/r]\n");
		return;
	}
	
	/* break it up */
	tokenize(args, 2, ' ', token);
	
	/* get the device */
	iface = interface_find_name(token[0]);
	if( iface == NULL ) {
		errormsg(0, "show_hdrcache", "Could not locate interface '%s'", token[0]);
		return;
	}

	if( (token[1] == NULL) || (*token[1] == 's') ) {	
		notice("-- Send Header Cache for Interface [%s] --\n", iface->name);
		/* ok, show the info */
		if( iface->cache_ptrs.send_hdrcache != NULL ) {
			if( iface->cache_ptrs.send_hdrcache->size > 0 ) {
				hdrcache_dump(iface->cache_ptrs.send_hdrcache);
			} else {
				notice("%s has no send header cache\n", iface->name);
			}
		} else {
			notice("No Cache!\n");
		}
	} else if( *token[1] == 'r' ) {
		notice("-- Recv Header Cache for Interface [%s] --\n", iface->name);
		if( iface->cache_ptrs.recv_hdrcache != NULL ) {
			if( iface->cache_ptrs.recv_hdrcache->size > 0 ) {
				hdrcache_dump(iface->cache_ptrs.recv_hdrcache);
			} else {
				notice("%s has no receive header cache.\n", iface->name);
			}
		} else {
			notice("No Cache!\n");
		}
	} else {
		notice("Unknown mode '%s'\n", token[1]);
	}
	
	/* done */		
}




/*
**	WHICH_COMMAND
*/
int which_command(COMMAND_ENTRY commands[], char *str)
{
	int		i, length, matches, value;
	
	/* remove warning */
	value = CMD_UNKNOWN;
	
	/* calc */
	length = strlen(str);
	
	/* check 'em */
	i = 0;
	matches = 0;
	while( 1 ) {
		/* is this the last command */
		if( commands[i].value == -1 ) {	
			break;
		}
		
		/* compare the strings */
		if( strncasecmp(commands[i].command, str, length) == 0 ) {
			/* a match */
			matches++;
			value = commands[i].value;
		}
		
		/* next one */
		i++;
	}
	
	/* did we have any matches? */
	if( matches == 0 ) {
		return(CMD_UNKNOWN);
	} else if( matches > 1 ) {
		return(CMD_AMBIGIOUS);
	} else {
		return(value);
	}
	
}
