/*
** ether_bsd.c - ethernet BSD style
**
** Keith Zaback (krz@wcug.wwu.edu) Oct 98
*/

#include "config.h"
#ifdef	FURY_ETHERNET
#ifdef	FURY_BSD

/* includes */
#include "ipxfury.h"
#include <net/if.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

/* globals */
extern GLOBALS	ifg;

/* defines */
#define FURY_MAX_BPF	4

/* filter "programs" -- Ethernet II frame types -- */
struct bpf_insn ether_ii_filter[] = {
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),		/* get 2-byte type field */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0x8137, 0, 1),	/* compare with 0x8137 */
	BPF_STMT(BPF_RET+BPF_K, (u_int)-1),		/* accept all the bytes */
	BPF_STMT(BPF_RET+BPF_K, 0),			/* drop this */
};
struct bpf_program ether_ii_program = {
	sizeof(ether_ii_filter) / sizeof(struct bpf_insn),
	ether_ii_filter,
};

/* 802.3 filter */
struct bpf_insn ether_8023_filter[] = {
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),		/* get type/length */
	BPF_JUMP(BPF_JMP+BPF_JGT+BPF_K, 1500, 0, 1),	/* larger than 1500 */
	BPF_STMT(BPF_RET+BPF_K, 0),			/* FAIL */
	BPF_STMT(BPF_ST, 0),				/* store ACC at m[0] */
	BPF_STMT(BPF_LDX+BPF_W+BPF_MEM, 0),		/* IDX = ACC */
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 16),		/* IPX length field */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_X, 0, 0, 1),	/* compare lengths */
	BPF_STMT(BPF_RET+BPF_K, (u_int)-1),		/* use it */
	BPF_STMT(BPF_RET+BPF_K, 0),			/* FAIL */
};
struct bpf_program ether_8023_program = {
	sizeof(ether_8023_filter) / sizeof(struct bpf_insn),
	ether_8023_filter,
};

/* 802.2 filter */
struct bpf_insn ether_8022_filter[] = {
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),		/* get type/length */
	BPF_JUMP(BPF_JMP+BPF_JGT+BPF_K, 1500, 0, 1),	/* larger than 1500 */
	BPF_STMT(BPF_RET+BPF_K, 0),			/* FAIL */
	BPF_STMT(BPF_ST, 0),				/* store ACC at m[0] */
	BPF_STMT(BPF_LDX+BPF_W+BPF_MEM, 0),		/* IDX = ACC */
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 19),		/* IPX length field */
	BPF_STMT(BPF_ALU+BPF_ADD+BPF_K, 3),		/* add 3 */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_X, 0, 0, 1),	/* compare lengths */
	BPF_STMT(BPF_RET+BPF_K, (u_int)-1),		/* use it */
	BPF_STMT(BPF_RET+BPF_K, 0),			/* FAIL */
};
struct bpf_program ether_8022_program = {
	sizeof(ether_8022_filter) / sizeof(struct bpf_insn),
	ether_8022_filter,
};


/*
**	ETHER_OPEN
*/
int ether_open(INTERFACE *edev)
{
	int			i;
	int			bpf;
	char			filename[32];	
	struct ifreq		ifr;
	struct bpf_program	*filter_prog;

	/* open /dev/bpfx */
	bpf = -1;
	for( i = 0; i < FURY_MAX_BPF; i++ ) {
		/* make filename */
		sprintf(filename, "/dev/bpf%d", i);
		
		/* open it */
		bpf = open(filename, O_RDWR, 0);
		if( bpf > 0 )
			break;
	}
	
	/* did we get one */
	if( bpf < 0 ) {
		errormsg(0, "ether_open", "could not open /dev/bpf");
		return(0);
	}
	
	/* yup.. so set it */
	notice("Using /dev/bpf%d\n", i);
	edev->fd = bpf;
	
	/* get the buffer size */
	if( ioctl(bpf, BIOCGBLEN, &edev->data.e.e_buffer_len) < 0 ) {
		errormsg(0, "ether_open", "Could not get buffer size!");
		close(bpf);
		return(0);
	}
	notice("Buffer Length %d bytes\n", edev->data.e.e_buffer_len);
	
	
	/* attach interface */
	memset(&ifr, 0, sizeof(struct ifreq));
	strcpy(ifr.ifr_name, edev->name);
	if( ioctl(bpf, BIOCSETIF, &ifr) < 0 ) {
		errormsg(0, "ether_open", "Failed to attach interface '%s'", edev->name);
		return(0);
	}
	notice("Attached interface %s\n", edev->name);
	
	/* set it to immediate mode.. so we get stuff right away */
	i = 1;
	if( ioctl(bpf, BIOCIMMEDIATE, &i) < 0 ) {
		errormsg(0, "ether_open", "Failed to set immediate mode");
		close(bpf);
		return(0);
	}
	
	/* set promisc mode */
	if( ioctl(bpf, BIOCPROMISC, NULL) < 0 ) {
		errormsg(0, "ether_open", "Could not set promiscuous mode");
		close(bpf);
		return(0);
	}
	
	/* set the filter based on frame type.. use II only for now */
	switch( edev->data.e.e_frametype ) {
		case FURY_FRAMETYPE_II:
			filter_prog = &ether_ii_program;
			break;
			
		case FURY_FRAMETYPE_802_3:
			filter_prog = &ether_8023_program;
			break;
		
		case FURY_FRAMETYPE_802_2:
			filter_prog = &ether_8022_program;
			break;
			
		default:
			errormsg(0, "ether_open", "Unsupported frame type (%d)", edev->data.e.e_frametype);
			close(bpf);
			return(0);
	}
	
	/* set the filter */
	if( ioctl(bpf, BIOCSETF, filter_prog) < 0 ) {
		errormsg(0, "ether_open", "Failed to set packet filter!");
		close(bpf);
		return(0);
	}

	/* create the buffers */
	if( ! create_buffer(&edev->recvbuf, FURY_ETHER_RECV_BUF_SIZE) ) {
		close(bpf);
		return(0);
	}
	
	if( ! create_buffer(&edev->sendbuf, FURY_ETHER_TRANS_BUF_SIZE) ) {
		close(bpf);
		return(0);
	}
	
	/* alloc the bpf receive buffer */
	edev->data.e.e_buffer = (BYTE *)malloc(edev->data.e.e_buffer_len);
	if( NULL == edev->data.e.e_buffer ) {
		/* failed */
		errormsg(0, "ether_open", "Could not allocate buffer (%d bytes)", edev->data.e.e_buffer_len);
		close(bpf);
		return(0);
	}
	
	/* set flags */
	edev->flags 	|= FURY_IF_VALID;
	edev->state	= FURY_STATE_INITIAL;
	
	/* did ok */
	return(1);
}

/*
**	ETHER_CLOSE
*/
int ether_close(INTERFACE *edev)
{

	/* only use valid interfaces */
	if( (edev->flags & FURY_IF_VALID) != FURY_IF_VALID ) {
		errormsg(0, "ether_close", "called with non-valid interface");
		return(0);
	}
	
	/* free the bpf buffer */
	if( NULL != edev->data.e.e_buffer ) {
		free(edev->data.e.e_buffer);
	}
	
	/* free the send and recv buffer */
	free_buffer(&edev->recvbuf);
	free_buffer(&edev->sendbuf);

	/* close the file */
	close(edev->fd);
	
	/* remove flags */
	edev->flags &= ~FURY_IF_VALID;
	
	/* did fine */
	return(1);
}



/*
**	ETHER_READ - we can get more than 1 packet.. 
*/
int ether_read(INTERFACE *edev)
{
	int			bytes_read;
	BYTE			*data, *etherframe;
	struct bpf_hdr		*bhdr;
	IPXHEADER		*ipxheader;
	
	/* read the data */
	bytes_read = read(edev->fd, edev->data.e.e_buffer, edev->data.e.e_buffer_len);

	if( bytes_read < 0 ) {
		errormsg(0, "ether_read", "failed to read()");
		return(0);
	}
	
	/* set starting ptr */
	data = edev->data.e.e_buffer;
	bhdr = (struct bpf_hdr *)data;

	/* now process the data */
	while( bytes_read > 0 ) {
		/* make sure we got all the data */
		if( bhdr->bh_caplen != bhdr->bh_datalen ) {
#if 1
			errormsg(0, "ether_read", "Fragmented packet! data: %d  cap: %d", bhdr->bh_datalen, bhdr->bh_caplen);
#endif	
			return(0);
		
		} else {
			/* get the ipx data */
			etherframe = (BYTE *)(data + bhdr->bh_hdrlen);
			ipxheader = ether_get_ipx(edev->data.e.e_frametype, etherframe);
			if( ipxheader != NULL ) {
			
				/* adjust the stats */
				edev->stats.recv_bytes_recent	+= bhdr->bh_caplen; 
				edev->stats.bytes_rcvd		+= bhdr->bh_caplen;
				
				/* check the port */
				if( check_port(ifg.watched_ports, ntohs(ipxheader->ipx_dest.port)) ) {
					/* add it to the buffer */
					if( ! buffer_insert(&edev->recvbuf, (BYTE *)ipxheader, ntohs(ipxheader->length)) ) {
						errormsg(0, "ether_read", "Could not add ipx frame to receive buffer! (%d bytes)", ntohs(ipxheader->length));
						return(0);
					}
				}
			}
		}
		
		/* move ptrs.. and adjust length */
		bytes_read -= BPF_WORDALIGN(bhdr->bh_hdrlen + bhdr->bh_caplen);
		data += BPF_WORDALIGN(bhdr->bh_hdrlen + bhdr->bh_caplen);
		bhdr = (struct bpf_hdr *)data;
		
	}

	/* did ok */	
	return(1);
}


/*
**	ETHER_FLUSH
*/
int ether_flush(INTERFACE *edev)
{
	/* do the ioctl to flush */
	ioctl(edev->fd, BIOCFLUSH, NULL);
	
	/* ok */
	return(1);
}




#endif	/* FURY_BSD */
#endif	/* FURY_ETHERNET */
